/*****************************************************************************
 * $Id: ix-objlist.c,v 1.1 2005/08/28 16:54:40 killabyte Exp $
 *
 * This module contains subroutines which examine the objects in memory and
 * extract important info about them. They also mantain a list of the objects
 * in memory which is used to iterate on these objects in a non-arch-dependant
 * way.
 *
 * ---------------------------------------------------------------------------
 * pDI-Tools - portable Dynamic Instrumentation Tools
 *   (C) 2004, 2005 Gerardo Garca Pea
 *   Programmed by Gerardo Garca Pea - Inspired on CEPBA DItools
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
 *   USA
 *
 *****************************************************************************/

#include<config.h>
#include<ebeif.h>
#include<files.h>
#include<log.h>
#include<pdiconfig.h>
#include<irix/elfops.h>

#define OINAME(x)             ((char *) ((x)->oi_pathname))

/*****************************************************************************
 * Module's private variables
 *
 *   archobjlist
 *     A pool of arch-dependant structures, where each object is directly
 *     associated with a PDI_ELFOBJ structure in the no arch-dependant pool.
 *
 *****************************************************************************/

static PDI_IRIX_ELFOBJ *archobjlist;

/*****************************************************************************
 * int _pdi_irix_newElfObj(PDI_ELFOBJ *eo, int objpool_index)
 *
 * Description:
 *   Alloc memory for a new PDI_IRIX_ELFOBJ structure and initialize the field
 *   'o_arch' in 'eo' with a pointer to it.
 *
 * Parameters:
 *   eo            - object associated with the new PDI_IRIX_ELFOBJ structure.
 *   objpool_index - pDI-Tools uses a static pool of PDI_ELFOBJ structures.
 *                   This index is the position of the structure 'eo' in this
 *                   pool.
 *
 * Returns:
 *   0 if succesful, -1 otherwise.
 *
 *****************************************************************************/

int _pdi_irix_newElfObj(PDI_ELFOBJ *eo, int objpool_index)
{
  eo->o_arch = &(archobjlist[objpool_index]);

  return 0;
}

/*****************************************************************************
 * void _pdi_irix_freeElfObj(PDI_ELFOBJ *eo, int objpool_index)
 *
 * Description:
 *   Free memory used by a PDI_IRIX_ELFOBJ structure.
 *
 * Parameters:
 *   eo            - object
 *   objpool_index - Index of structure 'eo' in PDI_ELFOBJ pool.
 *
 * Returns:
 *   nothing.
 *
 *****************************************************************************/

void _pdi_irix_freeElfObj(PDI_ELFOBJ *eo, int objpool_index)
{
  free(eo->o_name);
  eo->o_name = NULL;
}

/*****************************************************************************
 * static void initObject(PDI_ELFOBJ *eo, ElfW(Obj_Info) *oi)
 *
 * Description:
 *   Check a lot of arch-dependant information of object 'oi', and caches some
 *   values in structure 'eo'.
 *
 * Parameters:
 *   eo - A pointer to the PDI_ELFOBJ structure to initialize
 *   oi - ElfW(Obj_Info) (specific info about object)
 *
 * Returns:
 *   nothing.
 *
 *****************************************************************************/

#define IO_GET_PT(x,y)   case DT_##x: flags[f_##x]++; eo->o_arch->y = dyn->d_un.d_ptr + eo->o_arch->delta; break;
#define IO_GET_VL(x,y)   case DT_##x: flags[f_##x]++; eo->o_arch->y = dyn->d_un.d_val; break;

static int initObject(PDI_ELFOBJ *eo, ElfW(Obj_Info) *oi)
{
  ElfW(Dyn) *dyn;
  enum { f_STRTAB, f_STRSZ, f_SYMTAB,      f_SYMENT,
         f_PLTGOT, f_HASH,  f_MIPS_GOTSYM, f_MIPS_LOCAL_GOTNO };
  int flags[15];
  char tmp[PATH_MAX];

  /* Save a pointer to the ElfW(Obj_Info) structure and get the DSO filename */
  eo->o_arch->oi = oi;
  if(OINAME(oi) && strlen(OINAME(oi)) > 0)
  {
    if(_pdi_getCanonicalPath(OINAME(oi), tmp, 0))
    {
      _pdi_error(THIS, "Cannot canonicalize path '%s'", OINAME(oi));
      return -1;
    }
    if(!strlen(tmp))
      strcpy(tmp, OINAME(oi));
  } else
    *tmp = '\0';
  if((eo->o_name = malloc(strlen(tmp)+1)) == NULL)
  {
    _pdi_error(THIS, "No memory for string '%s'", tmp);
    return -1;
  }
  strcpy(eo->o_name, tmp);
  eo->o_alias = NULL;

  /* Calculta delta offset */
  eo->o_arch->delta = oi->oi_ehdr - oi->oi_orig_ehdr;

  /* Optimistically assume this object is ok */
  eo->o_arch->broken_dso = 0;

  /* Search and cache some required sections                                 */
  /* NOTE: Macros IO_GET_PT & IO_GET_VL get and fix the pointer/value stored */
  /*       in a dynamic tag and save it in some field of structure 'eo'.     */
  /*       IO_GET_PT is used to get a pointer, while IO_GET_VL is used to    */
  /*       read an ElfW(Word).                                               */
  memset(flags, 0, sizeof(flags));
  for(dyn = _pdi_irix_getFirstDynEntry(oi); dyn->d_tag != DT_NULL; dyn++)
    switch(dyn->d_tag)
    {
      IO_GET_PT(STRTAB,           stringTable)     /* mandatory */
      IO_GET_VL(STRSZ,            stringTableSize) /* mandatory */
      IO_GET_PT(SYMTAB,           symbolTable)     /* mandatory */
      IO_GET_VL(SYMENT,           symbolSize)      /* mandatory */
      IO_GET_PT(MIPS_LOCAL_GOTNO, localGotElems)   /* mandatory */
      IO_GET_PT(MIPS_GOTSYM,      gotSym)          /* mandatory */
      IO_GET_PT(HASH,             hashTable)       /* mandatory */
      IO_GET_PT(PLTGOT,           gotTable)        /* mandatory */
    }

  if(!flags[f_STRTAB]      || !flags[f_STRSZ]
  || !flags[f_SYMTAB]      || !flags[f_SYMENT]
  || !flags[f_HASH]        || !flags[f_PLTGOT]
  || !flags[f_MIPS_GOTSYM] || !flags[f_MIPS_LOCAL_GOTNO])
  {
    _pdi_error(THIS, "Object '%s' is broken. Cannot find:", OINAME(oi));
    if(!flags[f_STRTAB])           _pdi_error(THIS, "  - DT_STRTAB");
    if(!flags[f_STRSZ])            _pdi_error(THIS, "  - DT_STRSZ");
    if(!flags[f_SYMTAB])           _pdi_error(THIS, "  - DT_SYMTAB");
    if(!flags[f_SYMENT])           _pdi_error(THIS, "  - DT_SYMENT");
    if(!flags[f_HASH])             _pdi_error(THIS, "  - DT_HASH");
    if(!flags[f_PLTGOT])           _pdi_error(THIS, "  - DT_PLTGOT");
    if(!flags[f_MIPS_GOTSYM])      _pdi_error(THIS, "  - DT_MIPS_GOTSYM");
    if(!flags[f_MIPS_LOCAL_GOTNO]) _pdi_error(THIS, "  - DT_MIPS_LOCAL_GOTNO");
    eo->o_arch->broken_dso = 1;
  } else {
    /* Extract info about the hash table */
    eo->o_arch->n_bucket = *(((unsigned int *) eo->o_arch->hashTable) + 0);
    eo->o_arch->n_chain  = *(((unsigned int *) eo->o_arch->hashTable) + 1);
    eo->o_arch->l_bucket =   ((unsigned int *) eo->o_arch->hashTable) + 2 + 0;
    eo->o_arch->l_chain  =   ((unsigned int *) eo->o_arch->hashTable) + 2 + eo->o_arch->n_bucket;
  }

  return 0;
}

/*****************************************************************************
 * static int buildObjectList(void)
 *
 * Description:
 *   Builds the list of PDI_ELFOBJ objects in memory.
 *
 * Parameters:
 *   none.
 *
 * Returns:
 *   0 if succesful, != 0 if any error.
 *
 *****************************************************************************/

static int buildObjectList(void)
{
  ElfW(Obj_Info) *oi;
  PDI_ELFOBJ *eo;

  /* for each Object_Info, create a new PDI_ELFOBJ object */
  for(oi = __rld_obj_head; oi; oi = (ElfW(Obj_Info) *) oi->oi_next)
  {
    /* Alloc a PDI_ELFOBJ structure */
    if((eo = _pdi_ebe_newElfObj(NULL)) == NULL)
    {
      _pdi_debug(THIS, "Cannot alloc memory for a new PDI_ELFOBJ.");
      return -1;
    }

    _pdi_debug(THIS, "Add object '%s' to the object list.", OINAME(oi));
      
    /* Initialize it */
    if(initObject(eo, oi))
    {
      _pdi_error(THIS, "Cannot initialize object '%s'.", OINAME(oi));
      return -1;
    }
  }

  return 0;
}

/*****************************************************************************
 * static int initMainObject(void)
 *
 * Description:
 *   Search, initialize and set alias PDI_ALIAS_MAIN to the executable binary.
 *
 * Parameters:
 *   none.
 *
 * Returns:
 *   0 if succesful, != 0 if any error.
 *
 *****************************************************************************/

static int initMainObject(void)
{
  /* Make '_pdi_mainobj' point to the PDI_ELFOBJ describing the executable */
  /* binary, and set on it alias PDI_ALIAS_MAIN                            */
  /* NOTE: In Irix the first object of the object list is the executable   */
  _pdi_mainobj = _pdi_objlist;
  if(_pdi_ebe_setObjectAlias(_pdi_mainobj, PDI_ALIAS_MAIN))
  {
    _pdi_error(THIS, "Cannot assign alias " PDI_ALIAS_MAIN " to object '%s'.",
               _pdi_ebe_getObjectName(_pdi_mainobj));
    return -1;
  }

  return 0;
}

/*****************************************************************************
 * static int initLIBCObject(void)
 *
 * Description:
 *   Search, initialize and set alias PDI_ALIAS_LIBC to the C Shared Library.
 *
 *   NOTE: If cannot find 'libc.so.*' in the object list this function prints a
 *         warning, libc ausence is not an error.
 *
 * Parameters:
 *   none.
 *
 * Returns:
 *   0 if succesful, != 0 if any error.
 *
 *****************************************************************************/

static int initLIBCObject(void)
{
  PDI_ELFOBJ *eo;
  char *fname;
  
  /* Search in the object list this library */
  for(eo = _pdi_objlist; eo; eo = eo->o_next)
  {
    /* From the path select only the filename */
    for(fname = eo->o_name + strlen(eo->o_name);
        fname > eo->o_name && *(fname-1) != '/';
        fname--)
      ;
    if(!strncmp("libc.so", fname, strlen("libc.so")))
      break;
  }

  if(!eo)
  {
    _pdi_warning(THIS, "Cannot find 'libc.so.*' in the object list.");
    _pdi_libcobj = NULL;
    return 0;
  }

  /* Make a Direct Access (patent pending) to this object */
  _pdi_libcobj = eo;
  if(_pdi_ebe_setObjectAlias(_pdi_libcobj, PDI_ALIAS_LIBC))
  {
    _pdi_error(THIS, "Cannot assign alias " PDI_ALIAS_LIBC " to '%s'.",
               _pdi_pdiobj->o_name);
    return -1;
  }

  return 0;
}

/*****************************************************************************
 * static int initPDIObject(void)
 *
 * Description:
 *   Search and set alias PDI_ALIAS_PDI to pDI-Tools DSO.
 *
 *   NOTE: pDI-Tools can be linked statically, so is not an error to not find
 *         it. If pDI-Tools DSO (libpdi.so) cannot be found then '_pdi_mainobj'
 *         and '_pdi_pdiobj' are the same and a warning is printed.
 *
 * Parameters:
 *   none.
 *
 * Returns:
 *   0 if succesful, != 0 if any error.
 *
 *****************************************************************************/

static int initPDIObject(void)
{
  PDI_ELFOBJ *eo;
  char *fname;
  
  /* Search in the object list 'libpdi.so' (pDI-Tools DSO) */
  for(eo = _pdi_objlist; eo; eo = eo->o_next)
  {
    /* From the path select only the filename */
    for(fname = eo->o_name + strlen(eo->o_name);
        fname > eo->o_name && *(fname-1) != '/';
        fname--)
      ;
    if(!strcmp("libpdi.so", fname))
      break;
  }

  if(!eo)
  {
    _pdi_warning(THIS, "Cannot find 'libpdi.so' in the object list.");
    _pdi_warning(THIS, "Assuming '%s' as object '%s'.",
                 PDI_ALIAS_MAIN, PDI_ALIAS_PDI);
    _pdi_pdiobj = _pdi_mainobj;
    return 0;
  }

  /* Create a Direct Link (patent pending) to this object and aliase (TM) it */
  _pdi_pdiobj = eo;
  if(_pdi_ebe_setObjectAlias(_pdi_pdiobj, PDI_ALIAS_PDI))
  {
    _pdi_error(THIS, "Cannot assign alias " PDI_ALIAS_PDI " to '%s'.",
               _pdi_pdiobj->o_name);
    return -1;
  }

  return 0;
}

/*****************************************************************************
 * int _pdi_irix_initObjectList(void)
 *
 * Description:
 *   Initializes the arch dependant part of the object list, which is
 *   practically all the object list.
 *
 * Parameters:
 *   none.
 *
 * Returns:
 *   0 if succesful, != 0 if any error.
 *
 *****************************************************************************/

int _pdi_irix_initObjectList(void)
{
  /* Alloc memory for 'archobjlist' pool. It must have the same number */
  /* of entries as the 'objpool' pool of '/src/link/gestobj.c'.        */
  if((archobjlist = malloc(sizeof(PDI_IRIX_ELFOBJ) * PDICFG.max_objects)) == NULL)
  {
    _pdi_debug(THIS, "Cannot alloc memory for 'archobjlist' pool (%d bytes).",
               sizeof(PDI_IRIX_ELFOBJ) * PDICFG.max_objects);
    return -1;
  }
  _pdi_log(THIS, "Reserved %d bytes for 'archobjlist' table.",
           sizeof(PDI_IRIX_ELFOBJ) * PDICFG.max_objects);

  /* Build the object list (initializing it) */
  if(buildObjectList())
  {
    _pdi_debug(THIS, "Cannot initialize object list");
    return -1;
  }

  /* Search basic objects */
  if(initMainObject()) _pdi_debug(THIS, "Cannot find object " PDI_ALIAS_MAIN);
  if(initLIBCObject()) _pdi_debug(THIS, "Cannot find object " PDI_ALIAS_LIBC);
  if(initPDIObject())  _pdi_debug(THIS, "Cannot find object " PDI_ALIAS_PDI);

  return 0;
}

/*****************************************************************************
 * void _pdi_irix_finiObjectList(void)
 *
 * Description:
 *   Free all memory used by the arch dependant part of the object list.
 *
 * Parameters:
 *   none.
 *
 * Returns:
 *   nothing.
 *
 *****************************************************************************/

void _pdi_irix_finiObjectList(void)
{
  free(archobjlist);
  archobjlist = NULL;
}

/*****************************************************************************
 * int _pdi_irix_refreshObjectList(void)
 *
 * Description:
 *   Refreshes object list comparing the current object list with the runtime
 *   linker object list.
 *
 * Parameters:
 *   none.
 *
 * Returns:
 *   0 if succesful, != 0 if any error.
 *
 *****************************************************************************/

int _pdi_irix_refreshObjectList(void)
{
  ElfW(Obj_Info) *oi;
  PDI_ELFOBJ *eo, *next, *neo;

  _pdi_debug(THIS, "Removing DSO's not in memory.");
  /* check that all objects in our list are currently in memory */
  for(eo = _pdi_objlist; eo; eo = next)
  {
    /* remember the next 'eo' */
    next = eo->o_next;

    /* search in the list of linkmaps the object 'eo' */
    for(oi = __rld_obj_head; oi && eo->o_arch->oi != oi; oi = (ElfW(Obj_Info) *) oi->oi_next)
      ;

    /* if this object is not in memory send it to go jump in the digital river */
    if(!oi)
    {
      _pdi_debug(THIS, "Removing object '%s' from the object list.",
                 _pdi_ebe_getObjectName(eo));
      _pdi_ebe_freeElfObj(eo);
    }
  }

  _pdi_debug(THIS, "Adding new objects.");
  for(oi = __rld_obj_head, eo = _pdi_objlist; oi; oi = (ElfW(Obj_Info) *) oi->oi_next)
  {
    if(!eo || oi != eo->o_arch->oi)
    {
      _pdi_debug(THIS, "Adding '%s' to the object list.", OINAME(oi));
      
      if((neo = _pdi_ebe_newElfObj(eo)) == NULL)
      {
        _pdi_error(THIS, "Cannot alloc memory for a new PDI_ELFOBJ.");
        return -1;
      }
      
      initObject(neo, oi);
    } else
      eo = eo->o_next;
  }

  return 0;
}

